import _Object$keys from 'babel-runtime/core-js/object/keys';
import invariant from 'invariant';
import { Store } from '../Store';
import toMap from './toMap';

import { Iterable, Map } from 'immutable';

import { STORE_INSERT, STORE_REMOVE, STORE_FILTER, STORE_UPDATE } from '../constants/StoreConstants';

export default function createReducerForStore(store) {
  invariant(store && Store.isStore(store), 'Reducer: `store` is expected to be an Orchestra.Store.');

  var identifier = store.getIdentifier();
  var pre = store.getPre();
  var initial = store.createCollection();

  return function storeReducer() {
    var state = arguments.length <= 0 || arguments[0] === undefined ? initial : arguments[0];
    var action = arguments[1];

    if (identifier !== action.identifier) {
      return state;
    }

    var type = action.type;
    var payload = action.payload;
    var groupId = action.groupId;
    var selector = action.selector;

    switch (type) {
      case STORE_INSERT:
        {
          if (Map.isMap(payload)) {
            var item = pre(payload);
            if (!item) {
              return state;
            }

            var id = item.get('id');
            var _res = state.set(id, item);

            return groupId ? _res.addIdToGroup(groupId, id) : _res;
          }

          var res = state.withMutations(function (map) {
            // Deduping the incoming data by ids, since Immutable has a bug where keys
            // have to be unique while using mutable data.
            var track = {};

            payload.forEach(function (value) {
              var item = pre(value);
              if (!item) {
                return state;
              }

              var id = item.get('id');
              if (!track[id]) {
                track[id] = true;
                map.set(id, item);
              }
            });
          });

          return groupId ? res.addIdsToGroup(groupId, _Object$keys(track)) : res;
        }

      case STORE_REMOVE:
        {
          if (!payload) {
            return state;
          } else if (Map.isMap(payload)) {
            var _id = payload.get('id');
            return state.delete(_id);
          }

          return state.delete(payload);
        }

      case STORE_FILTER:
        {
          if (typeof selector !== 'function') {
            return state;
          }

          return state.filter(selector);
        }

      case STORE_UPDATE:
        {
          if (typeof selector !== 'function') {
            return state;
          }

          return state.map(function (x) {
            var item = pre(selector(x));
            return item ? item : x;
          });
        }

      default:
        return state;
    }
  };
}