'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.Collection = undefined;

var _classCallCheck2 = require('babel-runtime/helpers/classCallCheck');

var _classCallCheck3 = _interopRequireDefault(_classCallCheck2);

var _possibleConstructorReturn2 = require('babel-runtime/helpers/possibleConstructorReturn');

var _possibleConstructorReturn3 = _interopRequireDefault(_possibleConstructorReturn2);

var _inherits2 = require('babel-runtime/helpers/inherits');

var _inherits3 = _interopRequireDefault(_inherits2);

var _symbol = require('babel-runtime/core-js/symbol');

var _symbol2 = _interopRequireDefault(_symbol);

exports.default = createCollection;

var _invariant = require('invariant');

var _invariant2 = _interopRequireDefault(_invariant);

var _extendableImmutable = require('extendable-immutable');

var _immutable = require('immutable');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var groupsSymbol = (0, _symbol2.default)('groups');

var Collection = exports.Collection = function (_OrderedMap) {
  (0, _inherits3.default)(Collection, _OrderedMap);

  function Collection() {
    (0, _classCallCheck3.default)(this, Collection);
    return (0, _possibleConstructorReturn3.default)(this, _OrderedMap.apply(this, arguments));
  }

  Collection.of = function of(val) {
    return new Collection(val);
  };

  Collection.isCollection = function isCollection(val) {
    return val && val instanceof Collection;
  };

  Collection.prototype.__wrapImmutable = function __wrapImmutable() {
    var _OrderedMap$prototype;

    for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
      args[_key] = arguments[_key];
    }

    var res = (_OrderedMap$prototype = _OrderedMap.prototype.__wrapImmutable).call.apply(_OrderedMap$prototype, [this].concat(args));

    if (res.size) {
      res[groupsSymbol] = this[groupsSymbol] || {};
    } else {
      res[groupsSymbol] = {};
    }

    return res;
  };

  Collection.prototype.merge = function merge(x) {
    if (!this.size) {
      return x;
    }

    var res = _OrderedMap.prototype.merge.call(this, x);
    if (!x instanceof Collection) {
      return res;
    }

    var xGroups = x[groupsSymbol] || {};
    var groups = res[groupsSymbol] || {};

    for (var groupId in xGroups) {
      if (xGroups.hasOwnProperty(groupId)) {
        var _xGroup = xGroups[groupId] || new _immutable.OrderedSet();
        var _group = groups[groupId] || new _immutable.OrderedSet();

        groups[groupId] = _group.union(_xGroup);
      }
    }

    return res;
  };

  Collection.prototype.addIdToGroup = function addIdToGroup(groupId, id) {
    (0, _invariant2.default)(typeof groupId === 'string', 'Collection: `groupId` is expected to be an id (string).');
    (0, _invariant2.default)(typeof id === 'string', 'Collection: `id` is expected to be an id (string).');
    (0, _invariant2.default)(this.size && this.has(id), 'Collection: `id` is expected to be part of this collection.');

    var groups = this[groupsSymbol];

    var _group = groups[groupId] || new _immutable.OrderedSet();
    groups[groupId] = _group.add(id);

    return this;
  };

  Collection.prototype.addIdsToGroup = function addIdsToGroup(groupId, ids) {
    var _this2 = this;

    (0, _invariant2.default)(typeof groupId === 'string', 'Collection: `groupId` is expected to be an id (string).');
    (0, _invariant2.default)(_immutable.Set.isSet(ids) && ids.every(function (x) {
      return typeof x === 'string';
    }), 'Collection: `ids` is expected to be a Set containing ids (string).');
    (0, _invariant2.default)(this.size && ids.every(function (x) {
      return _this2.has(x);
    }), 'Collection: `ids` is expected to contain ids that are part of this collection.');

    var groups = this[groupsSymbol];

    var _group = groups[groupId] || new _immutable.OrderedSet();
    groups[groupId] = _group.union(ids);

    return this;
  };

  Collection.prototype.filterForGroup = function filterForGroup(groupId) {
    var _this3 = this;

    (0, _invariant2.default)(typeof groupId === 'string', 'Collection: `groupId` is expected to be an id (string).');

    var groups = this[groupsSymbol] || {};
    var _group = groups[groupId] || new _immutable.OrderedSet();

    var empty = this.clear();
    if (!_group.size) {
      return empty;
    }

    return empty.withMutations(function (map) {
      _group.forEach(function (id) {
        var item = _this3.get(id);
        if (item) {
          map.set(id, item);
        }
      });
    });
  };

  Collection.prototype.toString = function toString() {
    return this.__toString('Collection {', '}');
  };

  return Collection;
}(_extendableImmutable.OrderedMap);

function createCollection(obj) {
  return new Collection(obj);
}