'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.Orchestra = undefined;

var _extends3 = require('babel-runtime/helpers/extends');

var _extends4 = _interopRequireDefault(_extends3);

var _keys = require('babel-runtime/core-js/object/keys');

var _keys2 = _interopRequireDefault(_keys);

var _typeof2 = require('babel-runtime/helpers/typeof');

var _typeof3 = _interopRequireDefault(_typeof2);

var _classCallCheck2 = require('babel-runtime/helpers/classCallCheck');

var _classCallCheck3 = _interopRequireDefault(_classCallCheck2);

var _symbol = require('babel-runtime/core-js/symbol');

var _symbol2 = _interopRequireDefault(_symbol);

exports.default = createOrchestra;

var _invariant = require('invariant');

var _invariant2 = _interopRequireDefault(_invariant);

var _rxjs = require('@reactivex/rxjs');

var _Collection = require('./Collection');

var _Store = require('./Store');

var _combineStores = require('./util/combineStores');

var _combineStores2 = _interopRequireDefault(_combineStores);

var _isDispatcher = require('fluorine-lib/lib/util/isDispatcher');

var _isDispatcher2 = _interopRequireDefault(_isDispatcher);

var _immutable = require('immutable');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var resultCache = (0, _symbol2.default)('resultCache');

var Orchestra = exports.Orchestra = function () {
  Orchestra.isOrchestra = function isOrchestra(obj) {
    return (typeof obj === 'undefined' ? 'undefined' : (0, _typeof3.default)(obj)) === 'object' && obj instanceof Orchestra;
  };

  function Orchestra() {
    (0, _classCallCheck3.default)(this, Orchestra);

    for (var _len = arguments.length, stores = Array(_len), _key = 0; _key < _len; _key++) {
      stores[_key] = arguments[_key];
    }

    (0, _invariant2.default)(stores.length > 0 && stores.every(_Store.Store.isStore), 'Orchestra: constructor expects to receive Stores.');

    var _stores = stores.reduce(function (acc, store) {
      var identifier = store.identifier;


      if (acc[identifier]) {
        throw new Error('Orchestra: The identifier `' + identifier + '` is not unique.');
      }

      acc[identifier] = store;
      return acc;
    }, {});

    this[resultCache] = {};
    this.stores = _stores;
    this.externals = {};
  }

  Orchestra.prototype.addReducer = function addReducer(identifier, reducer) {
    (0, _invariant2.default)(typeof identifier === 'string', 'Orchestra: `identifier` is expected to be a string.');
    (0, _invariant2.default)(typeof reducer === 'function', 'Orchestra: `reducer` is expected to be a reducer function.');

    var externals = this.externals;
    var stores = this.stores;


    (0, _invariant2.default)(!stores.hasOwnProperty(identifier), 'Orchesta: The identifier `' + identifier + '` is already taken by a Store.');
    (0, _invariant2.default)(!externals.hasOwnProperty(identifier), 'Orchesta: The identifier `' + identifier + '` is not unique.');

    externals[identifier] = reducer;
    return this;
  };

  Orchestra.prototype.reduce = function reduce(dispatcher) {
    (0, _invariant2.default)((0, _isDispatcher2.default)(dispatcher), 'Orchestra: `dispatcher` is expected to be a Fluorine dispatcher.');
    if (this[resultCache][dispatcher.identifier]) {
      return this[resultCache][dispatcher.identifier];
    }

    var stores = this.stores;
    var externals = this.externals;


    var _externals = (0, _keys2.default)(externals).reduce(function (acc, key) {
      var external = externals[key];
      acc[key] = dispatcher.reduce(external);
      return acc;
    }, {});

    var _stores = {};

    var resolveStore = function resolveStore(store) {
      var _extends2;

      var visited = arguments.length <= 1 || arguments[1] === undefined ? {} : arguments[1];
      var identifier = store.identifier;


      if (visited[identifier]) {
        throw new Error('Orchestra: Failed to resolve circular dependency for identifier `' + identifier + '`.');
      }

      visited[identifier] = true;

      if (_stores[identifier]) {
        return _stores[identifier];
      }

      var reducer = store.getReducer();
      var dependencies = store.getDependencies();
      var post = store.getPost();

      var _dependencies = (0, _keys2.default)(dependencies).reduce(function (acc, dependency) {
        var _dependency = void 0;

        if (_externals[dependency]) {
          _dependency = _externals[dependency];
        } else if (_stores[dependency]) {
          _dependency = _stores[dependency];
        } else if (stores[dependency]) {
          _dependency = resolveStore(stores[dependency], visited);
        } else {
          throw new Error('Orchestra: Failed to resolve dependency for identifier `' + dependency + '`.');
        }

        acc[dependency] = _dependency;
        return acc;
      }, {});

      var _store = dispatcher.reduce(reducer);
      var res = (0, _combineStores2.default)((0, _extends4.default)({}, _dependencies, (_extends2 = {}, _extends2[identifier] = _store, _extends2))).map(function (deps) {
        // Resolve all dependencies
        var state = (0, _keys2.default)(dependencies).reduce(function (acc, dependencyIdentifier) {
          var dependencyState = deps[dependencyIdentifier];
          var _dependencies$depende = dependencies[dependencyIdentifier];
          var getter = _dependencies$depende.getter;
          var setter = _dependencies$depende.setter;

          // If getter exists we need to resolve dependencies per ids

          if (getter && typeof getter === 'function') {
            var _ret = function () {
              var missingIds = [];

              var nextState = acc.map(function (x) {
                var ids = getter(x);
                var result = undefined;

                if (typeof ids === 'string') {
                  result = dependencyState.get(ids);

                  if (!result) {
                    missingIds = missingIds.concat(ids);
                    return x;
                  }
                } else if (_immutable.Iterable.isIterable(ids) || Array.isArray(ids)) {
                  (0, _invariant2.default)(typeof ids.forEach === 'function', 'Orchestra: `ids` is expected to have a method `forEach`.');

                  var _store2 = stores[dependencyIdentifier];
                  var collection = _store2 ? _store2.createCollection() : _Collection.Collection;

                  result = collection.withMutations(function (map) {
                    ids.forEach(function (id) {
                      var item = dependencyState.get(id);
                      if (item === undefined) {
                        missingIds.push(id);
                      }

                      map.set(id, item);
                    });
                  });
                }

                return setter(x, result);
              });

              // Report missing items for ids
              var dependencyStore = stores[dependencyIdentifier];
              if (dependencyStore) {
                dependencyStore._missing(new _immutable.Set(missingIds), identifier);
              }

              return {
                v: nextState
              };
            }();

            if ((typeof _ret === 'undefined' ? 'undefined' : (0, _typeof3.default)(_ret)) === "object") return _ret.v;
          }

          // If there is no getter we just map over the items with the setter only
          return acc.map(function (x) {
            return setter(x, dependencyState);
          });
        }, deps[identifier]);

        return state.map(post);
      }).distinctUntilChanged().publishReplay(1).refCount();

      _stores[identifier] = res;

      return res;
    };

    for (var identifier in stores) {
      if (stores.hasOwnProperty(identifier)) {
        var store = stores[identifier];
        _stores[identifier] = resolveStore(store);
      }
    }

    this[resultCache][dispatcher.identifier] = _stores;
    return _stores;
  };

  return Orchestra;
}();

function createOrchestra() {
  for (var _len2 = arguments.length, stores = Array(_len2), _key2 = 0; _key2 < _len2; _key2++) {
    stores[_key2] = arguments[_key2];
  }

  return new (Function.prototype.bind.apply(Orchestra, [null].concat(stores)))();
}